@@ -33,8 +33,8 @@
 
 (define-vop (compare-and-swap-slot)
   (:args (object :scs (descriptor-reg))
-         (old :scs (descriptor-reg any-reg))
-         (new :scs (descriptor-reg any-reg)))
+         (old :scs (descriptor-reg any-reg zero))
+         (new :scs (descriptor-reg any-reg zero)))
   (:info name offset lowtag)
   (:ignore name)
   (:temporary (:sc non-descriptor-reg) lip)
@@ -247,8 +247,8 @@
 (define-vop (%compare-and-swap-symbol-value)
   (:translate %compare-and-swap-symbol-value)
   (:args (symbol :scs (descriptor-reg))
-         (old :scs (descriptor-reg any-reg))
-         (new :scs (descriptor-reg any-reg)))
+         (old :scs (descriptor-reg any-reg zero))
+         (new :scs (descriptor-reg any-reg zero)))
   (:results (result :scs (descriptor-reg any-reg) :from :load))
   #+sb-thread
   (:temporary (:sc any-reg) tls-index)
@@ -351,7 +351,7 @@
     (load-type type function (- fun-pointer-lowtag))
     (inst cmp type simple-fun-widetag)
     (inst b :eq SIMPLE-FUN)
-    (load-inline-constant lip '(:fixup closure-tramp :assembly-routine))
+    (load-asm-routine lip 'closure-tramp)
     SIMPLE-FUN
     (storew lip fdefn fdefn-raw-addr-slot other-pointer-lowtag)
     (storew function fdefn fdefn-fun-slot other-pointer-lowtag)
@@ -364,7 +364,7 @@
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 38
     (storew null-tn fdefn fdefn-fun-slot other-pointer-lowtag)
-    (load-inline-constant temp '(:fixup undefined-tramp :assembly-routine))
+    (load-asm-routine temp 'undefined-tramp)
     (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)))
 
 
@@ -392,8 +392,7 @@
       (inst ldr (32-bit-reg tls-index) (tls-index-of symbol))
       (inst cbnz (32-bit-reg tls-index) TLS-INDEX-VALID)
       (move alloc-tls-symbol symbol)
-      (load-inline-constant lr '(:fixup alloc-tls-index :assembly-routine))
-      (inst blr lr)
+      (invoke-asm-routine 'alloc-tls-index lr)
       TLS-INDEX-VALID
       (inst ldr alloc-tls-symbol (@ thread-tn tls-index))
       (inst stp alloc-tls-symbol tls-index
@@ -593,7 +592,7 @@
   (:args (object)
          (index)
          (old-value :scs (unsigned-reg))
-         (new-value :scs (unsigned-reg)))
+         (new-value :scs (unsigned-reg zero)))
   (:arg-types * tagged-num unsigned-num unsigned-num)
   (:results (result :scs (unsigned-reg) :from :load))
   (:result-types unsigned-num)
@@ -603,12 +602,37 @@
   (:args (object)
          (index)
          (old-value :scs (signed-reg))
-         (new-value :scs (signed-reg)))
+         (new-value :scs (signed-reg zero)))
   (:arg-types * tagged-num signed-num signed-num)
   (:results (result :scs (signed-reg) :from :load))
   (:result-types signed-num)
   (:translate %raw-instance-cas/signed-word))
 
+(define-vop (%instance-cas-v8.1 word-index-cas-v8.1)
+  (:policy :fast-safe)
+  (:translate %instance-cas)
+  (:variant instance-slots-offset instance-pointer-lowtag)
+  (:arg-types instance tagged-num * *))
+
+(define-vop (%raw-instance-cas/word-v8.1 %instance-cas-v8.1)
+  (:args (object)
+         (index)
+         (old-value :scs (unsigned-reg) :target result)
+         (new-value :scs (unsigned-reg zero)))
+  (:arg-types * tagged-num unsigned-num unsigned-num)
+  (:results (result :scs (unsigned-reg) :from (:argument 2)))
+  (:result-types unsigned-num)
+  (:translate %raw-instance-cas/word))
+
+(define-vop (%raw-instance-cas/signed-word-v8.1 %instance-cas-v8.1)
+  (:args (object)
+         (index)
+         (old-value :scs (signed-reg) :target result)
+         (new-value :scs (signed-reg zero)))
+  (:arg-types * tagged-num signed-num signed-num)
+  (:results (result :scs (signed-reg) :from (:argument 2)))
+  (:result-types signed-num)
+  (:translate %raw-instance-cas/signed-word))
 
 ;;;; Code object frobbing.
 
@@ -626,8 +650,7 @@
   (:temporary (:scs (non-descriptor-reg)) temp card)
   (:temporary (:sc non-descriptor-reg) pa-flag)
   (:generator 10
-    (load-inline-constant temp `(:fixup "gc_card_table_mask" :foreign-dataref))
-    (inst ldr temp (@ temp))
+    (load-foreign-symbol temp "gc_card_table_mask" :dataref t)
     (inst ldr (32-bit-reg temp) (@ temp)) ; 4-byte int
     (pseudo-atomic (pa-flag)
       #+immobile-space
@@ -667,8 +690,7 @@
       (inst lsr card object gencgc-card-shift)
       (inst and card card temp)
       ;; Load mark table base
-      (load-inline-constant temp `(:fixup "gc_card_mark" :foreign-dataref))
-      (inst ldr temp (@ temp))
+      (load-foreign-symbol temp "gc_card_mark" :dataref t)
       (inst ldr temp (@ temp))
       ;; Touch the card mark byte.
       (inst strb null-tn (@ temp card))
