@@ -124,18 +124,18 @@
 ;;; running full tilt will exhaust this faster, but it's still plenty enough.
 (defconstant sb-kernel::internal-time-bits 61)
 
-(defconstant most-positive-exactly-single-float-fixnum
-  (min (expt 2 single-float-digits) most-positive-fixnum))
-(defconstant most-negative-exactly-single-float-fixnum
-  (max (- (expt 2 single-float-digits)) most-negative-fixnum))
-(defconstant most-positive-exactly-double-float-fixnum
-  (min (expt 2 double-float-digits) most-positive-fixnum))
-(defconstant most-negative-exactly-double-float-fixnum
-  (max (- (expt 2 double-float-digits)) most-negative-fixnum))
+(defconstant most-positive-exactly-single-float-integer
+  (expt 2 single-float-digits))
+(defconstant most-negative-exactly-single-float-integer
+  (- (expt 2 single-float-digits)))
+(defconstant most-positive-exactly-double-float-integer
+  (expt 2 double-float-digits))
+(defconstant most-negative-exactly-double-float-integer
+  (- (expt 2 double-float-digits)))
 
 ;;;; Point where continuous area starting at dynamic-space-start bumps into
 ;;;; next space. Computed for genesis/constants.h, not used in Lisp.
-#+(and gencgc sb-xc-host)
+#+(and generational sb-xc-host)
 (defconstant max-dynamic-space-end
     (let ((stop (1- (ash 1 n-word-bits)))
           (start dynamic-space-start))
@@ -162,10 +162,15 @@
 ;;; which can be expressed in 8 bits.
 (defconstant short-header-max-words #x7fff)
 
-#+gencgc
-(defconstant max-conses-per-page
-  (floor (* gencgc-page-bytes n-byte-bits)
-         (1+ (* n-word-bytes 2 n-byte-bits)))) ; 1 extra bit per cons
+#+generational
+(progn
+  (defconstant gencgc-page-words (/ gencgc-page-bytes n-word-bytes))
+  ;; Preventing use of the last 2 words on a page ensures that we never
+  ;; "extend" an allocation region of conses from one page to the next page.
+  ;; All other things being equal, it's better for each page to start
+  ;; a contiguous block.
+  (defconstant max-conses-per-page
+    (1- (floor gencgc-page-words 2))))
 
 ;;; Amount to righ-shift an instance header to get the length.
 ;;; Similar consideration as above with regard to use of generation# byte.
@@ -184,6 +189,3 @@
 (defconstant gc-safepoint-trap-offset n-word-bytes)
 
 #+sb-xc-host (deftype sb-xc:fixnum () `(signed-byte ,n-fixnum-bits))
-
-#+gencgc
-(defconstant gencgc-page-words (/ gencgc-page-bytes n-word-bytes))
