@@ -16,7 +16,7 @@
 ;;; Simple TYPEP uses that don't have any standard predicate are
 ;;; translated into non-standard unary predicates.
 (defknown (fixnump bignump ratiop
-           short-float-p single-float-p double-float-p long-float-p
+           single-float-p double-float-p long-float-p
            complex-rational-p complex-float-p complex-single-float-p
            complex-double-float-p #+long-float complex-long-float-p
            complex-vector-p
@@ -162,7 +162,7 @@
 (defknown vector-sap ((simple-unboxed-array (*))) system-area-pointer
   (flushable))
 
-#+gencgc
+#+generational
 (defknown generation-of (t) (or (signed-byte 8) null) (flushable))
 
 ;;; WIDETAG-OF needs extra code to handle LIST and FUNCTION lowtags.
@@ -321,12 +321,16 @@
 (defknown make-weak-pointer (t) weak-pointer
   (flushable))
 
+;; This used to have a :derive-type but it can't now. Even though a weak vector
+;; happens to be a simple-vector (by default anyway), I don't plan to enable
+;; the "new" weak vector type (which is not an array) to be operated on by
+;; sequence functions. Maybe LENGTH and ELT but that's about it.
 (defknown make-weak-vector (index &key (:initial-element t)
                                        (:initial-contents t))
-  simple-vector (flushable)
-  :derive-type (lambda (call)
-                 (derive-make-array-type (first (combination-args call))
-                                         't nil nil nil call)))
+  weak-vector
+  (flushable))
+
+(defknown weak-vector-len (weak-vector) index (flushable))
 
 (defknown %make-complex (real real) complex
   (flushable movable))
@@ -606,16 +610,6 @@
     (values bignum-element-type bignum-element-type)
     (foldable flushable movable always-translatable))
 
-(defknown %lognot (bignum-element-type) bignum-element-type
-    (foldable flushable movable always-translatable))
-
-(defknown (%logand %logior %logxor) (bignum-element-type bignum-element-type)
-  bignum-element-type
-  (foldable flushable movable))
-
-(defknown %fixnum-to-digit (fixnum) bignum-element-type
-  (foldable flushable movable #-(or arm arm64) always-translatable))
-
 ;;; This takes three digits and returns the FLOOR'ed result of
 ;;; dividing the first two as a 2*digit-size integer by the third.
 (defknown %bigfloor (bignum-element-type bignum-element-type bignum-element-type)
@@ -735,14 +729,19 @@
 ;;; formerly in 'float-tran'
 
 (defknown %single-float (real) single-float
-  (movable foldable no-verify-arg-count))
+  (movable foldable unboxed-return))
 (defknown %double-float (real) double-float
-  (movable foldable no-verify-arg-count))
+  (movable foldable unboxed-return))
 
-(defknown bignum-to-float (bignum symbol) float
-  (movable foldable no-verify-arg-count))
-(defknown sb-kernel::float-ratio (ratio symbol) float
-  (movable foldable no-verify-arg-count))
+(defknown bignum-to-single-float (bignum) single-float
+  (movable foldable unboxed-return))
+(defknown bignum-to-double-float (bignum) double-float
+  (movable foldable unboxed-return))
+
+(defknown sb-kernel::double-float-ratio (ratio) double-float
+    (movable foldable unboxed-return))
+(defknown sb-kernel::single-float-ratio (ratio) single-float
+  (movable foldable unboxed-return))
 
 (defknown make-single-float ((signed-byte 32)) single-float
   (movable flushable))
@@ -844,3 +843,6 @@
 #+(or arm mips ppc sparc)
 (defknown sb-vm::+-modfx (integer integer) fixnum
           (movable foldable flushable always-translatable))
+
+(defknown sb-vm::%weakvec-ref (weak-pointer index) t (flushable))
+(defknown sb-vm::%weakvec-set (weak-pointer index t) (values) ())
