@@ -13,13 +13,13 @@
 #include <signal.h>
 
 #include "sbcl.h"
+#include "genesis/sap.h"
 #include "runtime.h"
 #include "os.h"
 #include "interrupt.h"
 #include "arch.h"
 #include "lispregs.h"
 #include "globals.h"
-#include "alloc.h"
 #include "breakpoint.h"
 #include "thread.h"
 #include "code.h"
