@@ -50,19 +50,6 @@ lispobj *current_control_frame_pointer;
 lispobj *current_binding_stack_pointer;
 #endif
 
-/* ARM + RISCV use ALLOCATION-POINTER, not dynamic_space_free_pointer */
-
-# if !(defined LISP_FEATURE_ARM || defined LISP_FEATURE_ARM64 || defined LISP_FEATURE_RISCV)
-/* The Object Formerly Known As current_dynamic_space_free_pointer */
-/* The ARM ports (32- and 64-bit) use a static-space lisp symbol for this pointer.
- * They also do not have a reg_ALLOC, so there is no ambiguity as to where the
- * pointer is obtained from: it is always in the static-space lisp symbol.
- * We do not copy that value back and forth to the C variable.
- * Moreover, because they do not have a reg_ALLOC, the pseudo-atomic flag
- * is not overlayed on the reg_ALLOC but is instead in another static-space
- * lisp symbol */
-lispobj *dynamic_space_free_pointer;
-#endif
 lispobj *read_only_space_free_pointer;
 lispobj *static_space_free_pointer;
 
@@ -71,21 +58,19 @@ lispobj *static_code_space_free_pointer;
 #endif
 
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
-lispobj *varyobj_free_pointer;
+lispobj *text_space_highwatermark;
 lispobj *fixedobj_free_pointer;
+lispobj ALIEN_LINKAGE_TABLE_SPACE_START;
 #endif
 os_vm_address_t anon_dynamic_space_start;
+// The end of immobile text mapped from disk, equivalently the starting address
+// of new objects handed out by the code allocator.
+lispobj* tlsf_mem_start; // meaningful only if immobile space
 
-#ifndef LISP_FEATURE_GENCGC /* GENCGC has its own way to record trigger */
-lispobj *current_auto_gc_trigger;
-#endif
-
-/* For cheneygc, this points to the start of the semi-space currently in use
- * (that will become the from_space when the next GC is done).
- * Gencgc defines it as DYNAMIC_SPACE_START via a C preprocessor macro. */
-#ifdef LISP_FEATURE_CHENEYGC
-lispobj *current_dynamic_space;
-#endif
+lispobj lisp_package_vector;
+// Tagged lisp pointer to a 'struct arena' (which is also a lisp DEFSTRUCT)
+// The chain terminates with NIL.
+lispobj arena_chain;
 
 void globals_init(void)
 {
@@ -95,11 +80,6 @@ void globals_init(void)
     current_control_frame_pointer = (lispobj *)0;
 #endif
 
-#ifndef LISP_FEATURE_GENCGC
-    /* no GC trigger yet */
-    current_auto_gc_trigger = NULL;
-#endif
-
 #ifndef LISP_FEATURE_SB_THREAD
 #if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)
     /* KLUDGE: x86oids always think they're in lisp code.  See the
