@@ -14,7 +14,7 @@
 #include "validate.h"
 #include "os.h"
 #include "lispregs.h"
-#include "signal.h"
+#include <signal.h>
 #include "pseudo-atomic.h"
 #include "interrupt.h"
 #include "interr.h"
@@ -48,7 +48,7 @@ os_context_insn(os_context_t *context)
         return *(unsigned int *)((unsigned)OS_CONTEXT_PC(context));
 }
 
-boolean
+bool
 arch_insn_with_bdelay_p(unsigned int insn)
 {
     switch (insn >> 26) {
@@ -249,16 +249,16 @@ arch_internal_error_arguments(os_context_t *context)
         return (unsigned char *)((unsigned)OS_CONTEXT_PC(context) + INSN_LEN);
 }
 
-boolean arch_pseudo_atomic_atomic(__attribute((unused)) os_context_t *context) {
-    return get_pseudo_atomic_atomic(get_sb_vm_thread());
+bool arch_pseudo_atomic_atomic(struct thread *thread) {
+    return get_pseudo_atomic_atomic(thread);
 }
 
-void arch_set_pseudo_atomic_interrupted(__attribute__((unused)) os_context_t *context) {
-    set_pseudo_atomic_interrupted(get_sb_vm_thread());
+void arch_set_pseudo_atomic_interrupted(struct thread *thread) {
+    set_pseudo_atomic_interrupted(thread);
 }
 
-void arch_clear_pseudo_atomic_interrupted(__attribute__((unused)) os_context_t *context) {
-    clear_pseudo_atomic_interrupted(get_sb_vm_thread());
+void arch_clear_pseudo_atomic_interrupted(struct thread *thread) {
+    clear_pseudo_atomic_interrupted(thread);
 }
 
 unsigned int
@@ -450,7 +450,7 @@ sigtrap_handler(int signal, siginfo_t *info, os_context_t *context)
     case 0x36: // #b110110 = TNE
         switch (code) {
         case 0: // FIXME: why is this not trap_PendingInterrupt ?
-          arch_clear_pseudo_atomic_interrupted(context);
+          arch_clear_pseudo_atomic_interrupted(get_sb_vm_thread());
           OS_CONTEXT_PC(context) += 4;
           return interrupt_handle_pending(context);
         case trap_InvalidArgCount:
@@ -497,7 +497,7 @@ arch_write_linkage_table_entry(int index, void *target_addr, int datap)
 {
     // allocate successive entries downward
     char *reloc_addr =
-        (char*)LINKAGE_TABLE_SPACE_END - (index + 1) * LINKAGE_TABLE_ENTRY_SIZE;
+        (char*)ALIEN_LINKAGE_TABLE_SPACE_END - (index + 1) * ALIEN_LINKAGE_TABLE_ENTRY_SIZE;
     *(unsigned int *)reloc_addr = (unsigned int)target_addr;
 }
 
