@@ -32,6 +32,9 @@
 #if defined(LISP_FEATURE_OS_PROVIDES_DLOPEN) && !defined(LISP_FEATURE_WIN32)
 # include <dlfcn.h>
 #endif
+#if defined LISP_FEATURE_UNIX && defined LISP_FEATURE_SOFT_CARD_MARKS
+#include "gc.h" // for find_page_index
+#endif
 
 /*
  * historically, this used sysconf to select the runtime page size
@@ -201,7 +204,7 @@ bool gc_managed_heap_space_p(lispobj addr)
 {
     if ((READ_ONLY_SPACE_START <= addr && addr < READ_ONLY_SPACE_END)
         || (STATIC_SPACE_START <= addr && addr < STATIC_SPACE_END)
-#if defined LISP_FEATURE_GENCGC
+#if defined LISP_FEATURE_GENERATIONAL
         || (DYNAMIC_SPACE_START <= addr &&
             addr < (DYNAMIC_SPACE_START + dynamic_space_size))
         || immobile_space_p(addr)
@@ -328,6 +331,18 @@ void* load_core_bytes_jit(int fd, os_vm_offset_t offset, os_vm_address_t addr, o
 
 #endif
 
+bool is_in_stack_space(lispobj ptr)
+{
+    struct thread *th;
+    for_each_thread(th) {
+        if ((th->control_stack_start <= (lispobj *)ptr) &&
+            (th->control_stack_end >= (lispobj *)ptr)) {
+            return 1;
+        }
+    }
+    return 0;
+}
+
 bool gc_managed_addr_p(lispobj addr)
 {
     struct thread *th;
@@ -372,7 +387,6 @@ char *copied_string(char *string)
 }
 
 #ifdef LISP_FEATURE_UNIX
-#include "gc-internal.h"
 void
 os_protect(os_vm_address_t address, os_vm_size_t length, os_vm_prot_t prot)
 {
