@@ -22,14 +22,12 @@
 #include "genesis/fdefn.h"
 #endif
 
-#ifdef LISP_FEATURE_GENCGC
-#include "gc-internal.h"
-#endif
+#include "gc.h"
 
 void os_init() {}
 
-os_vm_address_t os_validate(int attributes, os_vm_address_t addr, os_vm_size_t len,
-            int __attribute__((unused)) execute, int __attribute__((unused)) jit)
+os_vm_address_t os_alloc_gc_space(int __attribute__((unused)) space_id,
+                                  int attributes, os_vm_address_t addr, os_vm_size_t len)
 {
     int protection = attributes & IS_GUARD_PAGE ? OS_VM_PROT_NONE : OS_VM_PROT_ALL;
     attributes &= ~IS_GUARD_PAGE;
@@ -53,20 +51,14 @@ os_vm_address_t os_validate(int attributes, os_vm_address_t addr, os_vm_size_t l
     return addr;
 }
 
-void os_invalidate(os_vm_address_t addr, os_vm_size_t len)
-{
-    if(munmap((void*) addr, len) == -1)
-        perror("munmap");
-}
-
-#if defined LISP_FEATURE_GENCGC
+#if defined LISP_FEATURE_GENERATIONAL
 
 void
 sigsegv_handler(int signal, siginfo_t *info, os_context_t *context)
 {
     void* fault_addr = (void*)info->si_addr;
 
-    if (gencgc_handle_wp_violation(fault_addr)) return;
+    if (gencgc_handle_wp_violation(context, fault_addr)) return;
 
     if (!handle_guard_page_triggered(context, fault_addr))
             lisp_memory_fault_error(context, fault_addr);
