@@ -60,8 +60,6 @@
 #include "unaligned.h"
 
 #include "gc.h"
-#include "gc-private.h"
-#include "gencgc-internal.h"
 #include <wincrypt.h>
 #include <stdarg.h>
 #include <string.h>
